import fracnetics as fn 
import gymnasium as gym
import statistics
import time 
from gymnasium.wrappers import RecordVideo

start = time.perf_counter()
env = gym.make("LunarLander-v3")
# initializing population
pop = fn.Population(
    seed=42,
    ni=10_000,
    jn=1,
    jnf=8,
    pn=4,
    pnf=4,
    fractalJudgment=False
)

minFeatures = [ -2.5, -2.5, -10, -10, -6.2831855, -10, -0, -0 ] 
maxFeatures = [ 2.5, 2.5, 10, 10, 6.2831855, 10, 1, 1 ] 
pop.setAllNodeBoundaries(minFeatures,maxFeatures)

fitnessProgess = []
for g in range(100):
  pop.gymnasium(
          env,
          dMax=10,
          penalty=2,
          maxSteps=500,
          maxConsecutiveP=10,
          worstFitness=-200
          )
  pop.tournamentSelection(2,1)
  pop.callEdgeMutation(0.03, 0.03)
  pop.crossover(0.05)
  pop.callBoundaryMutationUniform(0.05)
  #pop.callBoundaryMutationFractal(0.05, minFeatures, maxFeatures)
  pop.callAddDelNodes(minFeatures,maxFeatures)
  maxFitness = pop.bestFit
  print(f"Generation: {g} | maxFitness: {maxFitness} | meanFitness: {pop.meanFitness}")
  fitnessProgess.append(maxFitness)
  #env.render()

pop.individuals[-1].fitness
print(f"Start Node: {pop.individuals[-1].startNode.edges}")
for node in pop.individuals[-1].innerNodes:
  print(f"Type: {node.type} | Function: {node.f} Edges: {node.edges} | Boundaries: {node.boundaries}")


print("Validation")
env = gym.make("LunarLander-v3")
validationResults = []
for v in range(10):
    pop.gymnasium(
          env,
          dMax=10,
          penalty=2,
          maxSteps=500,
          maxConsecutiveP=10,
          worstFitness=-200
          )
    validationResults.append(pop.bestFit)
print(statistics.mean(validationResults))

env = gym.make("LunarLander-v3", render_mode="rgb_array")
env = RecordVideo(env, video_folder="videos")
pop.individuals = [pop.individuals[-1]]
pop.gymnasium(
      env,
      dMax=10,
      penalty=1,
      maxSteps=500,
      maxConsecutiveP=5,
      worstFitness=-200)

env.close()



import pickle

def storePopulation(obj, filename):
    """
    Store a population object in a file using pickle serialization.

    Parameters
    ----------
    obj : Population
        The population object to store.
    filename : str
        Path or filename where the population should be saved.

    Returns
    -------
    None
        Prints a confirmation message upon successful storage.
    """
    with open(filename, 'wb') as f:
        pickle.dump(obj, f)
    print(f"Population successfully stored in '{filename}'.")


def loadPopulation(filename):
    """
    Load a previously stored population object from a file.

    Parameters
    ----------
    filename : str
        Path or filename of the stored population.

    Returns
    -------
    Population
        The deserialized population object.
    """
    with open(filename, 'rb') as f:
        obj = pickle.load(f)
    return obj



import fracnetics 
import sys

def test_population_init():
    try:
        seed = 42
        ni = 10
        jn = 5
        jnf = 3
        pn = 2
        pnf = 1
        fractal_judgment = True

        pop = fracnetics.Population(
            seed=seed,
            ni=ni,
            jn=jn,
            jnf=jnf,
            pn=pn,
            pnf=pnf,
            fractalJudgment=fractal_judgment
        )

        pop.ni
        pop.meanFitness
        pop.individuals[0]

        print("✅ initializing population")

    except Exception as e:
        print("❌ error initializing the population:")
        print(e)
        sys.exit(1)

if __name__ == "__main__":
    test_population_init()




# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import os
import sys
sys.path.insert(0, os.path.abspath('../../fracnetics'))

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'Fracnetics'
author = 'Fabian Köhnke'
copyright = '2025, Fabian Köhnke'
autodoc_typehints = 'description'
napoleon_google_docstring = True
napoleon_numpy_docstring = True

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.napoleon',
    'sphinx.ext.viewcode',
    'sphinx_autodoc_typehints',
    'breathe',
    'sphinx_gallery.gen_gallery',
]

# ---- Breathe (C++ Integration) ----
breathe_projects = {
    "Fracnetics": "../cpp/xml"
}
breathe_default_project = "Fracnetics"


templates_path = ['_templates']
exclude_patterns = []

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'pydata_sphinx_theme'
html_static_path = ['_static']

# ---- sphinx-gallery ----
sphinx_gallery_conf = {
    'examples_dirs': '../../examples',   # Python examples
    'gallery_dirs': 'auto_examples',
    'filename_pattern': r'/example_',
}

html_theme_options = {
    "icon_links": [
        {
            "name": "Examples",
            "url": "examples.html",
            "icon": "fa-solid fa-book-open",
            "type": "fontawesome",
        },
    ],
    "logo": {
        "text": "🦾 Fracnetics",
    },
}




import fracnetics as fn
import pandas as pd 
import sys

def test_population_function():

    # reading data 
    data = pd.read_csv("data/IRIS.csv")
    X = data.iloc[:,1:5].values
    y = data.iloc[:,5].values
    minF = data.iloc[:,1:5].min().values
    maxF = data.iloc[:,1:5].max().values

    # initializing population
    pop = fn.Population(
        seed=42,
        ni=10,
        jn=5,
        jnf=3,
        pn=3,
        pnf=2,
        fractalJudgment=True
    )
    
    try:
        pop.setAllNodeBoundaries(minF,maxF)
    except Exception as e:
        print("❌ error in pop.setAllNodeBoundaries()")
        print(e)
        sys.exit(1)

    print("✅ pop.setAllNodeBoundaries()") 

    try:
        fit1 = pop.individuals[0].fitness
        pop.accuracy(X,y,10,2)
        fit2 = pop.individuals[0].fitness
        if fit1 == fit2:
            print("no difference of fitnes after applying callFitness()")
            sys.exit(1)
    except Exception as e:
        print("❌ error in pop.callFitness()")
        print(e)
        sys.exit(1)

    print("✅ pop.callFitness()") 

    try:
        pop.tournamentSelection(2,1)
    except Exception as e:
        print("❌ error in pop.tournamentSelection()")
        print(e)
        sys.exit(1)

    print("✅ pop.tournamentSelection()") 
    
    try:
        edges = []
        for i in range(len(pop.individuals[0].innerNodes)):
            edges.append(pop.individuals[0].innerNodes[i].edges) 
        pop.callEdgeMutation(0.5,0)
        counter = 0
        for i in range(len(pop.individuals[0].innerNodes)):
            if edges[i] != pop.individuals[0].innerNodes[i].edges:
                counter += 1
        if counter == 0:
            print("❌ no changes by callEdgeMutation()!")
            sys.exit(1)
 
    except Exception as e:
        print("❌ error in pop.callEdgeMutation()")
        print(e)
        sys.exit(1)

    print("✅ pop.callEdgeMutation()") 

    def testBoundaryMutation(testFunc, name):
        try:
            boundaries = []
            for i in range(len(pop.individuals[0].innerNodes)):
                boundaries.append(pop.individuals[0].innerNodes[i].boundaries) 
            testFunc()
            counter = 0
            for i in range(len(pop.individuals[0].innerNodes)):
                if boundaries[i] != pop.individuals[0].innerNodes[i].boundaries:
                    counter += 1
            if counter == 0:
                print(f"❌ no changes by {name}!")
                sys.exit(1)

        except Exception as e:
            print(f"❌ error in pop.{name}")
            print(e)
            sys.exit(1)

        print(f"✅ pop.{name}") 

    testBoundaryMutation(lambda: pop.callBoundaryMutationUniform(1), "callBoundaryMutationUniform()")
    testBoundaryMutation(lambda: pop.callBoundaryMutationNormal(0.5,0.2), "callBoundaryMutationNormal()")
    testBoundaryMutation(lambda: pop.callBoundaryMutationNetworkSizeDependingSigma(0.5,0.2), "callBoundaryMutationNetworkSizeDependingSigma()")
    testBoundaryMutation(lambda: pop.callBoundaryMutationEdgeSizeDependingSigma(0.5,0.2), "callBoundaryMutationEdgeSizeDependingSigma()")
    testBoundaryMutation(lambda: pop.callBoundaryMutationFractal(0.8,minF,maxF), "callBoundaryMutationFractal()")
    
    try:
        edges = []
        for i in range(len(pop.individuals[0].innerNodes)):
            edges.append(pop.individuals[0].innerNodes[i].edges) 
        pop.crossover(0.5)
        counter = 0
        for i in range(len(pop.individuals[0].innerNodes)):
            if edges[i] != pop.individuals[0].innerNodes[i].edges:
                counter += 1
        if counter == 0:
            print("❌ no changes by crossover()!")
            sys.exit(1)

    except Exception as e:
        print(f"❌ error in pop.crossover()")
        print(e)
        sys.exit(1)

    try:
        pop.callAddDelNodes(minF,maxF)
    except Exception as e:
        print(f"❌ error in pop.callAddDelNodes()")
        print(e)
        sys.exit(1)

if __name__ == "__main__":
    test_population_function()

